/*
  Prop
  Manipulate properties of objects
  JBS 15 June 1994
*/

#ifndef _PROP_

#define _PROP_

#include "types.h"

void prop_getprop(word, word);
void prop_put_prop(word, word, word);
void prop_get_next_prop(word, word);
void prop_get_prop_addr(word, word);
void prop_get_p_len(word);

void load_word_array(word, word);
void load_byte_array(word, word);
void save_word_array(word, word, word);
void save_byte_array(word, word, word);

#endif
